@@ -9,11 +9,12 @@ import itertools
 import math
 import numbers
 import random
+import re
 import warnings
 
 from sentinels import NOTHING
 import six
-from six import moves
+from six import moves, raise_from
 
 from mongomock import command_cursor
 from mongomock import filtering
@@ -21,10 +22,12 @@ from mongomock import helpers
 from mongomock import OperationFailure
 
 try:
-    from bson import decimal128
+    from bson import Regex, decimal128
     decimal_support = True
+    _RE_TYPES = (helpers.RE_TYPE, Regex)
 except ImportError:
     decimal_support = False
+    _RE_TYPES = (helpers.RE_TYPE)
 
 _random = random.Random()
 
@@ -73,6 +76,7 @@ control_flow_operators = [
 ]
 projection_operators = ['$map', '$let', '$literal']
 date_operators = [
+    '$dateFromString',
     '$dateToString',
     '$dayOfMonth',
     '$dayOfWeek',
@@ -119,6 +123,7 @@ string_operators = [
     '$substrCP',
     '$toLower',
     '$toUpper',
+    '$trim',
 ]
 comparison_operators = [
     '$cmp',
@@ -224,10 +229,9 @@ class _Parser(object):
                 return self._handle_string_operator(k, v)
             if k in type_convertion_operators:
                 return self._handle_type_convertion_operator(k, v)
-            if k in boolean_operators + \
-                    text_search_operators + \
-                    projection_operators + \
-                    object_operators:
+            if k in boolean_operators:
+                return self._handle_boolean_operator(k, v)
+            if k in text_search_operators + projection_operators + object_operators:
                 raise NotImplementedError(
                     "'%s' is a valid operation but it is not supported by Mongomock yet." % k)
             if k.startswith('$'):
@@ -268,6 +272,20 @@ class _Parser(object):
             return helpers.get_value_by_dot(self._doc_dict, expression[1:], can_generate_array=True)
         return expression
 
+    def _handle_boolean_operator(self, operator, values):
+        if operator == '$and':
+            return all([self._parse_to_bool(value) for value in values])
+        if operator == '$or':
+            return any(self._parse_to_bool(value) for value in values)
+        if operator == '$not':
+            return not self._parse_to_bool(values)
+        # This should never happen: it is only a safe fallback if something went wrong.
+        raise NotImplementedError(  # pragma: no cover
+            "Although '%s' is a valid boolean operator for the "
+            'aggregation pipeline, it is currently not implemented'
+            ' in Mongomock.' % operator
+        )
+
     def _handle_arithmetic_operator(self, operator, values):
         if operator == '$abs':
             return abs(self.parse(values))
@@ -362,6 +380,22 @@ class _Parser(object):
         if operator == '$concat':
             parsed_list = list(self.parse_many(values))
             return None if None in parsed_list else ''.join([str(x) for x in parsed_list])
+        if operator == '$split':
+            if len(values) != 2:
+                raise OperationFailure('split must have 2 items')
+            try:
+                string = self.parse(values[0])
+                delimiter = self.parse(values[1])
+            except KeyError:
+                return None
+
+            if string is None or delimiter is None:
+                return None
+            if not isinstance(string, six.string_types):
+                raise TypeError('split first argument must evaluate to string')
+            if not isinstance(delimiter, six.string_types):
+                raise TypeError('split second argument must evaluate to string')
+            return string.split(delimiter)
         if operator == '$substr':
             if len(values) != 3:
                 raise OperationFailure('substr must have 3 items')
@@ -384,6 +418,66 @@ class _Parser(object):
                 raise OperationFailure('strcasecmp must have 2 items')
             a, b = str(self.parse(values[0])), str(self.parse(values[1]))
             return 0 if a == b else -1 if a < b else 1
+        if operator == '$regexMatch':
+            if not isinstance(values, dict):
+                raise OperationFailure(
+                    '$regexMatch expects an object of named arguments but found: %s' % type(values))
+            for field in ('input', 'regex'):
+                if field not in values:
+                    raise OperationFailure("$regexMatch requires '%s' parameter" % field)
+            unknown_args = set(values) - {'input', 'regex', 'options'}
+            if unknown_args:
+                raise OperationFailure(
+                    '$regexMatch found an unknown argument: %s' % list(unknown_args)[0])
+
+            try:
+                input_value = self.parse(values['input'])
+            except KeyError:
+                return False
+            if not isinstance(input_value, six.string_types):
+                raise OperationFailure("$regexMatch needs 'input' to be of type string")
+
+            try:
+                regex_val = self.parse(values['regex'])
+            except KeyError:
+                return False
+            options = None
+            for option in values.get('options', ''):
+                if option not in 'imxs':
+                    raise OperationFailure(
+                        '$regexMatch invalid flag in regex options: %s' % option)
+                re_option = getattr(re, option.upper())
+                if options is None:
+                    options = re_option
+                else:
+                    options |= re_option
+            if isinstance(regex_val, str):
+                if options is None:
+                    regex = re.compile(regex_val)
+                else:
+                    regex = re.compile(regex_val, options)
+            elif 'options' in values and regex_val.flags:
+                raise OperationFailure(
+                    "$regexMatch: regex option(s) specified in both 'regex' and 'option' fields")
+            elif isinstance(regex_val, helpers.RE_TYPE):
+                if options and not regex_val.flags:
+                    regex = re.compile(regex_val.pattern, options)
+                elif regex_val.flags & ~(re.I | re.M | re.X | re.S):
+                    raise OperationFailure(
+                        '$regexMatch invalid flag in regex options: %s' % regex_val.flags)
+                else:
+                    regex = regex_val
+            elif isinstance(regex_val, _RE_TYPES):
+                # bson.Regex
+                if regex_val.flags & ~(re.I | re.M | re.X | re.S):
+                    raise OperationFailure(
+                        '$regexMatch invalid flag in regex options: %s' % regex_val.flags)
+                regex = re.compile(regex_val.pattern, regex_val.flags or options)
+            else:
+                raise OperationFailure("$regexMatch needs 'regex' to be of type string or regex")
+
+            return bool(regex.search(input_value))
+
         # This should never happen: it is only a safe fallback if something went wrong.
         raise NotImplementedError(  # pragma: no cover
             "Although '%s' is a valid string operator for the aggregation "
@@ -568,10 +662,10 @@ class _Parser(object):
             elif isinstance(parsed, str):
                 try:
                     decimal_value = decimal128.Decimal128(parsed)
-                except decimal.InvalidOperation:
-                    raise OperationFailure(
+                except decimal.InvalidOperation as err:
+                    raise_from(OperationFailure(
                         "Failed to parse number '%s' in $convert with no onError value:"
-                        'Failed to parse string to decimal' % parsed)
+                        'Failed to parse string to decimal' % parsed), err)
             elif isinstance(parsed, datetime.datetime):
                 epoch = datetime.datetime.utcfromtimestamp(0)
                 string_micro_seconds = str((parsed - epoch).total_seconds() * 1000).split('.')[0]
@@ -947,10 +1041,10 @@ def _handle_bucket_stage(in_collection, unused_database, options):
     def _get_default_bucket():
         try:
             return options['default']
-        except KeyError:
-            raise OperationFailure(
+        except KeyError as err:
+            raise_from(OperationFailure(
                 '$bucket could not find a matching branch for '
-                'an input, and no default was specified.')
+                'an input, and no default was specified.'), err)
 
     def _get_bucket_id(doc):
         """Get the bucket ID for a document.
@@ -1227,6 +1321,14 @@ def _handle_facet_stage(in_collection, database, options):
     return [out_collection_by_pipeline]
 
 
+def _handle_match_stage(in_collection, database, options):
+    spec = helpers.patch_datetime_awareness_in_document(options)
+    return [
+        doc for doc in in_collection
+        if filtering.filter_applies(spec, helpers.patch_datetime_awareness_in_document(doc))
+    ]
+
+
 _PIPELINE_HANDLERS = {
     '$addFields': _handle_add_fields_stage,
     '$bucket': _handle_bucket_stage,
@@ -1243,7 +1345,7 @@ _PIPELINE_HANDLERS = {
     '$listLocalSessions': None,
     '$listSessions': None,
     '$lookup': _handle_lookup_stage,
-    '$match': lambda c, d, o: [doc for doc in c if filtering.filter_applies(o, doc)],
+    '$match': _handle_match_stage,
     '$merge': None,
     '$out': _handle_out_stage,
     '$planCacheStats': None,
@@ -1269,11 +1371,11 @@ def process_pipeline(collection, database, pipeline, session):
         for operator, options in six.iteritems(stage):
             try:
                 handler = _PIPELINE_HANDLERS[operator]
-            except KeyError:
-                raise NotImplementedError(
+            except KeyError as err:
+                raise_from(NotImplementedError(
                     '%s is not a valid operator for the aggregation pipeline. '
                     'See http://docs.mongodb.org/manual/meta/aggregation-quick-reference/ '
-                    'for a complete list of valid operators.' % operator)
+                    'for a complete list of valid operators.' % operator), err)
             if not handler:
                 raise NotImplementedError(
                     "Although '%s' is a valid operator for the aggregation pipeline, it is "
