@@ -9,7 +9,7 @@ import numbers
 import operator
 import re
 from sentinels import NOTHING
-from six import iteritems, iterkeys, string_types, PY3
+from six import iteritems, string_types, PY3
 try:
     from types import NoneType
 except ImportError:
@@ -170,7 +170,14 @@ class _Filterer(object):
             return False
         if not isinstance(query, dict):
             raise OperationFailure('$elemMatch needs an Object')
-        return any(self.apply(query, item) for item in doc_val)
+        for item in doc_val:
+            try:
+                if self.apply(query, item):
+                    return True
+            except OperationFailure:
+                if self.apply({'field': query}, {'field': item}):
+                    return True
+        return False
 
     def _all_op(self, doc_val, search_val):
         if isinstance(doc_val, list) and doc_val and isinstance(doc_val[0], list):
@@ -424,11 +431,6 @@ def _combine_regex_options(search):
         return search_copy
 
     if isinstance(search['$regex'], _RE_TYPES):
-        keys = [k for k in iterkeys(search) if k in {'$regex', '$options'}]
-        if keys == ['$options', '$regex']:
-            raise NotImplementedError(
-                'Do not use compiled regular expressions with $options until '
-                'https://jira.mongodb.org/browse/SERVER-38621 is solved.')
         if isinstance(search['$regex'], RE_TYPE):
             search_copy['$regex'] = re.compile(
                 search['$regex'].pattern, search['$regex'].flags | options)
